import os
import numpy as np
import pandas as pd
import seaborn as sns

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

from crabnet.utils.utils import CONSTANTS


# %%
cons = CONSTANTS()

for benchmark in [True, False]:
    df = pd.read_csv("metrics/all_metrics.csv")
    drop = ["n_total", "n_val", "prop_train", "prop_val", "prop_test"]
    df = df.drop(columns=drop, axis=0)
    df.index = df["property"].values

    if benchmark:
        convert = cons.benchmark_names_dict
        data_dir = "data/benchmark_data/"
        properties = os.listdir(data_dir)

    else:
        convert = cons.matbench_names_dict
        data_dir = "data/matbench/"
        drop = ["expt_is_metal", "glass", "mp_is_metal"]
        properties = [p[:-4] for p in os.listdir(data_dir) if p[:-4] not in drop]

    data_values = df.loc[properties].values

    # %%
    def color_scale_r2(value, min_val, max_val):
        n_colors = 100
        pal = sns.light_palette((210, 80, 50), input="husl", n_colors=n_colors * 1.0)
        diff = value - min_val
        idx = int((n_colors - 1) * (diff) / (max_val - min_val))
        color = pal[idx]
        return color

    def color_scale_mae(value, min_val, max_val):
        n_colors = 100
        pal = sns.diverging_palette(240, 10, n=n_colors * 1.0)
        diff = value - min_val
        idx = int((n_colors - 1) * (diff) / (max_val - min_val))
        color = pal[idx]

        if value < 0:
            color = "#fab098"

        return color

    # %%

    if benchmark:
        fig, ax = plt.subplots(figsize=(13, 9))
    else:
        fig, ax = plt.subplots(figsize=(13, 5))

    cell_h = 1
    cell_w = 1
    cell_width = 7
    left_bound = 14
    n_columns = 6

    for i, array in enumerate(data_values):

        denominator = 75  # add buffer to color scale bounds "max + std/demoniator"
        r2_max = np.nanmax(array[1:n_columns]) + array[n_columns + 4] / denominator
        r2_min = np.nanmin(array[1:n_columns]) - array[n_columns + 4] / denominator

        if i == 0:
            for j in range(1, n_columns):
                rect = Rectangle(
                    [j * cell_width, len(data_values)],
                    cell_width * cell_w,
                    1 * cell_h,
                    facecolor="w",
                    edgecolor="k",
                )
                ax.add_patch(rect)
                rect = Rectangle(
                    [j * cell_width, len(data_values)],
                    cell_width * cell_w,
                    1 * cell_h,
                    facecolor="w",
                    edgecolor="k",
                )
                ax.add_patch(rect)
                rect = Rectangle(
                    [j * cell_width, len(data_values)],
                    cell_width * cell_w,
                    1 * cell_h,
                    facecolor="w",
                    edgecolor="k",
                )
                ax.add_patch(rect)

        for j, value in enumerate(array):
            # print(value)
            # j = j
            if j == 0:
                rect = Rectangle(
                    [j - left_bound, i],
                    (left_bound + cell_width) * cell_w,
                    1 * cell_h,
                    facecolor="w",
                    edgecolor="k",
                )
                ax.add_patch(rect)

            if j >= 1 and j < n_columns:
                # print(array[0], j)
                value = float(value)
                if np.isnan(value) or np.isnan(r2_min) or np.isnan(r2_max):
                    print("nan values found")
                    rect = Rectangle(
                        [j * cell_width, len(data_values) - i - 1],
                        cell_width * cell_w,
                        1 * cell_h,
                        facecolor="gray",
                        edgecolor="k",
                    )
                else:
                    if j in [j * cell_w for j in range(len(array))]:

                        # if j == n_columns-1:
                        if value > r2_max or value < r2_min:
                            rect = Rectangle(
                                [j * cell_width, len(data_values) - i - 1],
                                cell_width * cell_w,
                                1 * cell_h,
                                facecolor="gray",
                                edgecolor="k",
                            )
                        else:
                            rect = Rectangle(
                                [j * cell_width, len(data_values) - i - 1],
                                cell_width * cell_w,
                                1 * cell_h,
                                facecolor=color_scale_mae(value, r2_min, r2_max),
                                edgecolor="k",
                            )
                ax.add_patch(rect)

            if j == 0:
                eformat = False
                if value == "aflow__agl_thermal_expansion_300K":
                    eformat = True
                plt.text(
                    j - left_bound + (0.5 * cell_w),
                    len(data_values) - (i + (0.5 * cell_h)),
                    f"{convert[value]}",
                    verticalalignment="center",
                    horizontalalignment="left",
                )

            elif j >= 1 and j < n_columns:
                if eformat:
                    plt.text(
                        j * cell_width + (1 * cell_width / 2),
                        len(data_values) - (i + (0.5 * cell_h)),
                        f"{value:0.2e}",
                        verticalalignment="center",
                        horizontalalignment="center",
                    )
                else:
                    plt.text(
                        j * cell_width + (1 * cell_width / 2),
                        len(data_values) - (i + (0.5 * cell_h)),
                        f"{value:0.3f}",
                        verticalalignment="center",
                        horizontalalignment="center",
                    )

    plt.xlim(0 - left_bound, cell_width * n_columns)
    plt.ylim(0, len(data_values) * cell_h + 1)

    plt.tick_params(
        axis="both",
        which="both",
        bottom=False,
        top=False,
        left=False,
        right=False,
        labelbottom=False,
        labelleft=False,
    )

    if benchmark:
        plt.text(
            -left_bound + (cell_width + left_bound) / 2,
            len(data_values) * cell_h + 0.5,
            "Extended Properties",
            verticalalignment="center",
            horizontalalignment="center",
        )
    else:
        plt.text(
            -left_bound + (cell_width + left_bound) / 2,
            len(data_values) * cell_h + 0.5,
            "MatBench Properties",
            verticalalignment="center",
            horizontalalignment="center",
        )

    plt.text(
        1 * cell_width + (1 * cell_width / 2),
        len(data_values) * cell_h + 0.5,
        "Roost",
        verticalalignment="center",
        horizontalalignment="center",
    )

    plt.text(
        2 * cell_width + (1 * cell_width / 2),
        len(data_values) * cell_h + 0.5,
        "CrabNet",
        verticalalignment="center",
        horizontalalignment="center",
    )

    plt.text(
        3 * cell_width + (1 * cell_width / 2),
        len(data_values) * cell_h + 0.5,
        "HotCrab",
        verticalalignment="center",
        horizontalalignment="center",
    )

    plt.text(
        4 * cell_width + (1 * cell_width / 2),
        len(data_values) * cell_h + 0.5,
        "ElemNet",
        verticalalignment="center",
        horizontalalignment="center",
    )

    plt.text(
        5 * cell_width + (1 * cell_width / 2),
        len(data_values) * cell_h + 0.5,
        "RF",
        verticalalignment="center",
        horizontalalignment="center",
    )

    outpath = os.path.join("figures/")

    if benchmark:
        save_name = os.path.join(outpath, "Table2_benchmark_metrics.png")
    else:
        save_name = os.path.join(outpath, "Table2_matbench_metrics.png")

    plt.savefig(save_name, dpi=300, bbox_inches="tight")
    plt.show()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['crabnet',
 'crabnet.data',
 'crabnet.data.benchmark_data',
 'crabnet.data.benchmark_data.CritExam__Ed',
 'crabnet.data.benchmark_data.CritExam__Ef',
 'crabnet.data.benchmark_data.OQMD_Bandgap',
 'crabnet.data.benchmark_data.OQMD_Energy_per_atom',
 'crabnet.data.benchmark_data.OQMD_Formation_Enthalpy',
 'crabnet.data.benchmark_data.OQMD_Volume_per_atom',
 'crabnet.data.benchmark_data.aflow__Egap',
 'crabnet.data.benchmark_data.aflow__ael_bulk_modulus_vrh',
 'crabnet.data.benchmark_data.aflow__ael_debye_temperature',
 'crabnet.data.benchmark_data.aflow__ael_shear_modulus_vrh',
 'crabnet.data.benchmark_data.aflow__agl_thermal_conductivity_300K',
 'crabnet.data.benchmark_data.aflow__agl_thermal_expansion_300K',
 'crabnet.data.benchmark_data.aflow__energy_atom',
 'crabnet.data.benchmark_data.mp_bulk_modulus',
 'crabnet.data.benchmark_data.mp_e_hull',
 'crabnet.data.benchmark_data.mp_elastic_anisotropy',
 'crabnet.data.benchmark_data.mp_mu_b',
 'crabnet.data.benchmark_data.mp_shear_modulus',
 'crabnet.data.element_properties',
 'crabnet.data.matbench',
 'crabnet.data.materials_data',
 'crabnet.data.materials_data.elasticity',
 'crabnet.data.materials_data.example_materials_property',
 'crabnet.utils']

package_data = \
{'': ['*'],
 'crabnet': ['tests/*'],
 'crabnet.data': ['matbench_cv/*',
                  'matbench_cv/castelli/*',
                  'matbench_cv/dielectric/*',
                  'matbench_cv/elasticity_log10(G_VRH)/*',
                  'matbench_cv/elasticity_log10(K_VRH)/*',
                  'matbench_cv/expt_gap/*',
                  'matbench_cv/expt_is_metal/*',
                  'matbench_cv/glass/*',
                  'matbench_cv/jdft2d/*',
                  'matbench_cv/mp_e_form/*',
                  'matbench_cv/mp_gap/*',
                  'matbench_cv/mp_is_metal/*',
                  'matbench_cv/phonons/*',
                  'matbench_cv/steels_yield/*']}

install_requires = \
['numpy',
 'pandas',
 'scikit-learn',
 'scipy',
 'matplotlib',
 'seaborn',
 'tqdm',
 'joblib',
 'psutil']

setup(name='crabnet',
      version='1.0.6',
      description='Predict materials properties using only the composition information.',
      author=None,
      author_email='"Sterling G. Baird" <sterling.baird@utah.edu>, Anthony Yu-Tung Wang <aytwang@alumni.uwaterloo.ca>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.7,<3.10',
     )



import os

import pandas as pd
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt
import seaborn as sns
from collections import OrderedDict, defaultdict

import torch
from torch.optim.optimizer import Optimizer
from torch.utils.data import Dataset, DataLoader
from torch import nn

from crabnet.utils.composition import (
    generate_features,
    _element_composition,
)

from sklearn.preprocessing import StandardScaler, Normalizer

import json

plt.rcParams.update({"font.size": 16})

RNG_SEED = 42
torch.manual_seed(RNG_SEED)
np.random.seed(RNG_SEED)

# %%
fig_dir = r"figures/Classics/"
data_type_torch = torch.float32
data_type_np = np.float32


# %%
class CONSTANTS:
    def __init__(self):
        self.crab_red = "#f2636e"
        self.dense_blue = "#2c2cd5"
        self.colors = list(sns.color_palette("Set1", n_colors=7, desat=0.5))

        self.markers = [
            "o",
            "x",
            "s",
            "^",
            "D",
            "P",
            "1",
            "2",
            "3",
            "4",
            "p",
            "*",
            "h",
            "H",
            "+",
            "d",
            "|",
            "_",
        ]

        self.eps = ["oliynyk", "jarvis", "mat2vec", "onehot", "magpie", "random_200"]

        self.benchmark_props = [
            "aflow__ael_bulk_modulus_vrh",
            "aflow__ael_debye_temperature",
            "aflow__ael_shear_modulus_vrh",
            "aflow__agl_thermal_conductivity_300K",
            "aflow__agl_thermal_expansion_300K",
            "aflow__Egap",
            "aflow__energy_atom",
            "CritExam__Ed",
            "CritExam__Ef",
            "mp_bulk_modulus",
            "mp_elastic_anisotropy",
            "mp_e_hull",
            "mp_mu_b",
            "mp_shear_modulus",
            "OQMD_Bandgap",
            "OQMD_Energy_per_atom",
            "OQMD_Formation_Enthalpy",
            "OQMD_Volume_per_atom",
        ]

        self.benchmark_names = [
            "AFLOW Bulk modulus",
            "AFLOW Debye temperature",
            "AFLOW Shear modulus",
            "AFLOW Thermal conductivity",
            "AFLOW Thermal expansion",
            "AFLOW Band gap",
            "AFLOW Energy per atom",
            "Bartel Decomposition (Ed)",
            "Bartel Formation (Ef)",
            "MP Bulk modulus",
            "MP Elastic anisotropy",
            "MP Energy above convex hull",
            "MP Magnetic moment",
            "MP Shear modulus",
            "OQMD Band gap",
            "OQMD Energy per atom",
            "OQMD Formation enthalpy",
            "OQMD Volume per atom",
        ]

        self.matbench_props = [
            "castelli",
            "dielectric",
            "elasticity_log10(G_VRH)",
            "elasticity_log10(K_VRH)",
            "expt_gap",
            "expt_is_metal",
            "glass",
            "jdft2d",
            "mp_e_form",
            "mp_gap",
            "mp_is_metal",
            "phonons",
            "steels_yield",
        ]

        self.matbench_names = [
            "Castelli perovskites",
            "Refractive index",
            "Shear modulus (log10)",
            "Bulk modulus (log10)",
            "Experimental band gap",
            "Experimental metallicity",
            "Experimental glass formation",
            "DFT Exfoliation energy",
            "MP Formation energy",
            "MP Band gap",
            "MP Metallicity",
            "Phonon peak",
            "Steels yield",
        ]

        self.benchmark_names_dict = dict(
            zip(self.benchmark_props, self.benchmark_names)
        )
        self.matbench_names_dict = dict(zip(self.matbench_props, self.matbench_names))

        self.mb_units_dict = {
            "castelli": "eV/unit cell",
            "dielectric": "unitless",
            "elasticity_log10(G_VRH)": "log(GPa)",
            "elasticity_log10(K_VRH)": "log(GPa)",
            "expt_gap": "eV",
            "expt_is_metal": "binary",
            "glass": "binary",
            "jdft2d": "meV/atom",
            "mp_e_form": "eV/atom",
            "mp_gap": "eV",
            "mp_is_metal": "binary",
            "phonons": "$cm^{−1}$",
            "steels_yield": "MPa",
        }

        self.bm_units_dict = {
            "aflow__ael_bulk_modulus_vrh": None,
            "aflow__ael_debye_temperature": None,
            "aflow__ael_shear_modulus_vrh": None,
            "aflow__agl_thermal_conductivity_300K": None,
            "aflow__agl_thermal_expansion_300K": None,
            "aflow__Egap": None,
            "aflow__energy_atom": None,
            "CritExam__Ed": None,
            "CritExam__Ef": None,
            "mp_bulk_modulus": None,
            "mp_elastic_anisotropy": None,
            "mp_e_hull": None,
            "mp_mu_b": None,
            "mp_shear_modulus": None,
            "OQMD_Bandgap": None,
            "OQMD_Energy_per_atom": None,
            "OQMD_Formation_Enthalpy": None,
            "OQMD_Volume_per_atom": None,
        }

        self.mp_units_dict = {
            "energy_atom": "eV/atom",
            "ael_shear_modulus_vrh": "GPa",
            "ael_bulk_modulus_vrh": "GPa",
            "ael_debye_temperature": "K",
            "Egap": "eV",
            "agl_thermal_conductivity_300K": "W/m*K",
            "agl_log10_thermal_expansion_300K": "1/K",
        }

        self.mp_sym_dict = {
            "energy_atom": "$E_{atom}$",
            "ael_shear_modulus_vrh": "$G$",
            "ael_bulk_modulus_vrh": "$B$",
            "ael_debye_temperature": "$\\theta_D$",
            "Egap": "$E_g$",
            "agl_thermal_conductivity_300K": "$\\kappa$",
            "agl_log10_thermal_expansion_300K": "$\\alpha$",
        }

        self.classification_list = ["mp_is_metal", "expt_is_metal", "glass"]

        self.classic_models_dict = {
            "Ridge": "Ridge",
            "SGDRegressor": "SGD",
            "ExtraTreesRegressor": "ExtraTrees",
            "RandomForestRegressor": "RF",
            "AdaBoostRegressor": "AdaBoost",
            "GradientBoostingRegressor": "GradBoost",
            "KNeighborsRegressor": "kNN",
            "SVR": "SVR",
            "lSVR": "lSVR",
        }

        # fmt: off
        self.atomic_symbols = [ "None", "H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne",
                               "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca", "Sc",
                               "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga" "Ge",
                               "As", "Se", "Br", "Kr", "Rb", "Sr", "Y", "Zr", "Nb", "Mo", "Tc",
                               "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe",
                               "Cs", "Ba", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb",
                               "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W", "Re", "Os",
                               "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr",
                               "Ra", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf",
                               "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt",
                               "Ds", "Rg", "Cn", "Nh", "Fl", "Mc", "Lv", "Ts", "Og"]
        # fmt: on

        self.idx_symbol_dict = {(i): sym for i, sym in enumerate(self.atomic_symbols)}


# %%
def get_cbfv(data, elem_prop="oliynyk", scale=False):
    """
    Loads the compound csv file and featurizes it, then scales the features
    using StandardScaler.

    Parameters
    ----------
    path : str
        DESCRIPTION.
    elem_prop : str, optional
        DESCRIPTION. The default is 'oliynyk'.

    Returns
    -------
    X_scaled : TYPE
        DESCRIPTION.
    y : TYPE
        DESCRIPTION.
    formula : TYPE
        DESCRIPTION.

    """
    if type(data) is str:
        df = pd.read_csv(data, keep_default_na=False, na_values=[""])
    else:
        df = data
    if "formula" not in df.columns.values.tolist():
        df["formula"] = df["cif_id"].str.split("_ICSD").str[0]
    # elem_prop = 'mat2vec'
    # elem_prop = 'oliynyk'
    mini = False
    # mini = True
    X, y, formula, skipped = generate_features(df, elem_prop, mini=mini)
    if scale:
        # scale each column of data to have a mean of 0 and a variance of 1
        scaler = StandardScaler()
        # normalize each row in the data
        normalizer = Normalizer()

        X_scaled = scaler.fit_transform(X)
        X_scaled = pd.DataFrame(
            normalizer.fit_transform(X_scaled),
            columns=X.columns.values,
            index=X.index.values,
        )

        return X_scaled, y, formula, skipped
    else:
        return X, y, formula, skipped


# %%
def BCEWithLogitsLoss(output, log_std, target):
    loss = nn.functional.binary_cross_entropy_with_logits(output, target)
    return loss


def RobustL1(output, log_std, target):
    """
    Robust L1 loss using a lorentzian prior. Allows for estimation
    of an aleatoric uncertainty.
    """
    absolute = torch.abs(output - target)
    loss = np.sqrt(2.0) * absolute * torch.exp(-log_std) + log_std
    return torch.mean(loss)


def RobustL2(output, log_std, target):
    """
    Robust L2 loss using a gaussian prior. Allows for estimation
    of an aleatoric uncertainty.
    """
    squared = torch.pow(output - target, 2.0)
    loss = 0.5 * squared * torch.exp(-2.0 * log_std) + log_std
    return torch.mean(loss)


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


# %%
class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.float):
            return float(obj)
        return json.JSONEncoder.default(self, obj)


def count_gs_param_combinations(d):
    cnt_dict = OrderedDict({})
    # array = []
    if isinstance(d, (list)) and not isinstance(d, (bool)):
        return len(d), cnt_dict
    elif isinstance(d, (int, float, complex)) and not isinstance(d, (bool)):
        return 1, cnt_dict
    elif isinstance(d, (bool)) or isinstance(d, (str)):
        return 1, cnt_dict
    elif d is None:
        return 1, cnt_dict
    elif isinstance(d, (dict, OrderedDict)):
        keys = d.keys()
        for k in keys:
            array = []
            subd = d[k]
            array.append(count_gs_param_combinations(subd)[0])
            cnt = np.prod(array)
            cnt_dict[k] = cnt
        return np.prod(list(cnt_dict.values())), cnt_dict
    return cnt, cnt_dict


# %%
class Scaler:
    def __init__(self, data):
        self.data = torch.as_tensor(data)
        self.mean = torch.mean(self.data)
        self.std = torch.std(self.data)

    def scale(self, data):
        data = torch.as_tensor(data)
        data_scaled = (data - self.mean) / self.std
        return data_scaled

    def unscale(self, data_scaled):
        data_scaled = torch.as_tensor(data_scaled)
        data = data_scaled * self.std + self.mean
        return data

    def state_dict(self):
        return {"mean": self.mean, "std": self.std}

    def load_state_dict(self, state_dict):
        self.mean = state_dict["mean"]
        self.std = state_dict["std"]


class DummyScaler:
    def __init__(self, data):
        self.data = torch.as_tensor(data)
        self.mean = torch.mean(self.data)
        self.std = torch.std(self.data)

    def scale(self, data):
        return torch.as_tensor(data)

    def unscale(self, data_scaled):
        return torch.as_tensor(data_scaled)

    def state_dict(self):
        return {"mean": self.mean, "std": self.std}

    def load_state_dict(self, state_dict):
        self.mean = state_dict["mean"]
        self.std = state_dict["std"]


# %%
class EDMDataset(Dataset):
    """
    Get X and y from EDM dataset.
    """

    def __init__(self, dataset, n_comp):
        self.data = dataset
        self.n_comp = n_comp

        self.X = np.array(self.data[0])
        self.y = np.array(self.data[1])
        self.formula = np.array(self.data[2])

        self.shape = [(self.X.shape), (self.y.shape), (self.formula.shape)]

    def __str__(self):
        string = f"EDMDataset with X.shape {self.X.shape}"
        return string

    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, idx):
        X = self.X[idx, :, :]
        y = self.y[idx]
        formula = self.formula[idx]

        X = torch.as_tensor(X, dtype=data_type_torch)
        y = torch.as_tensor(y, dtype=data_type_torch)

        return (X, y, formula)


def get_edm(
    data, elem_prop="mat2vec", n_elements="infer", inference=False, verbose=True
):
    """
    Build a element descriptor matrix.

    Parameters
    ----------
    data: str or DataFrame
        Filepath to data or DataFrame.
    elem_prop : str, optional
        DESCRIPTION. The default is 'oliynyk'.

    Returns
    -------
    X_scaled : TYPE
        DESCRIPTION.
    y : TYPE
        DESCRIPTION.
    formula : TYPE
        DESCRIPTION.

    """
    # fmt: off
    all_symbols = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na',
                   'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc',
                   'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga',
                   'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb',
                   'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb',
                   'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
                   'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',
                   'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl',
                   'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa',
                   'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
                   'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg',
                   'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']
    # fmt: on

    # mat_prop = 'phonons'
    # i = 0
    # path = rf'data\matbench_cv\{mat_prop}\test{i}.csv'
    if type(data) is str:
        df = pd.read_csv(data, keep_default_na=False, na_values=[""])
    else:
        df = data
    if "formula" not in df.columns.values.tolist():
        df["formula"] = df["cif_id"].str.split("_ICSD").str[0]

    df.loc[:, "count"] = [
        len(_element_composition(form)) for form in df.loc[:, "formula"]
    ]
    # df = df[df["count"] != 1]  # drop pure elements
    if not inference:
        df = df.groupby(by="formula").mean().reset_index()  # mean of duplicates

    list_ohm = [OrderedDict(_element_composition(form)) for form in df["formula"]]
    list_ohm = [
        OrderedDict(sorted(mat.items(), key=lambda x: -x[1])) for mat in list_ohm
    ]

    y = df["target"].values.astype(data_type_np)
    formula = df["formula"].values
    if n_elements == "infer":
        n_elements = 16

    edm_array = np.zeros(
        shape=(len(list_ohm), n_elements, len(all_symbols) + 1), dtype=data_type_np
    )
    elem_num = np.zeros(shape=(len(list_ohm), n_elements), dtype=data_type_np)
    elem_frac = np.zeros(shape=(len(list_ohm), n_elements), dtype=data_type_np)
    for i, comp in enumerate(
        tqdm(list_ohm, desc="Generating EDM", unit="formulae", disable=not verbose)
    ):
        for j, (elem, count) in enumerate(list_ohm[i].items()):
            if j == n_elements:
                # Truncate EDM representation to n_elements
                break
            try:
                edm_array[i, j, all_symbols.index(elem) + 1] = count
                elem_num[i, j] = all_symbols.index(elem) + 1
            except ValueError:
                print(f"skipping composition {comp}")

    # Scale features
    for i in range(edm_array.shape[0]):
        frac = edm_array[i, :, :].sum(axis=-1) / (edm_array[i, :, :].sum(axis=-1)).sum()
        elem_frac[i, :] = frac

    if n_elements == 16:
        n_elements = np.max(np.sum(elem_frac > 0, axis=1, keepdims=True))
        elem_num = elem_num[:, :n_elements]
        elem_frac = elem_frac[:, :n_elements]

    elem_num = elem_num.reshape(elem_num.shape[0], elem_num.shape[1], 1)
    elem_frac = elem_frac.reshape(elem_frac.shape[0], elem_frac.shape[1], 1)
    out = np.concatenate((elem_num, elem_frac), axis=1)

    return out, y, formula


# %%
class EDM_CsvLoader:
    """
    Parameters
    ----------
    data: str or DataFrame
        name of csv file containing cif and properties or DataFrame
    csv_val: str
        name of csv file containing cif and properties
    val_frac: float, optional (default=0.75)
        train/val ratio if val_file not given
    batch_size: float, optional (default=64)
        Step size for the Gaussian filter
    random_state: int, optional (default=123)
        Random seed for sampling the dataset. Only used if validation data is
        not given.
    shuffle: bool (default=True)
        Whether to shuffle the datasets or not
    """

    def __init__(
        self,
        data,
        batch_size=64,
        num_workers=1,
        random_state=0,
        shuffle=True,
        pin_memory=True,
        n_elements=6,
        inference=False,
        verbose=True,
    ):
        self.data = data
        self.main_data = list(
            get_edm(
                self.data,
                elem_prop="mat2vec",
                n_elements=n_elements,
                inference=inference,
                verbose=verbose,
            )
        )
        self.n_train = len(self.main_data[0])
        self.n_elements = self.main_data[0].shape[1] // 2

        self.batch_size = batch_size
        self.pin_memory = pin_memory
        self.shuffle = shuffle
        self.random_state = random_state

    def get_data_loaders(self, inference=False):
        """
        Input the dataset, get train test split
        """
        shuffle = not inference  # don't shuffle data when inferencing
        pred_dataset = EDMDataset(self.main_data, self.n_elements)
        pred_loader = DataLoader(
            pred_dataset,
            batch_size=self.batch_size,
            pin_memory=self.pin_memory,
            shuffle=shuffle,
        )
        return pred_loader


# %%
class Lamb(Optimizer):
    r"""Implements Lamb algorithm.
    It has been proposed in `Large Batch Optimization for Deep Learning:
        Training BERT in 76 minutes`_.
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        adam (bool, optional): always use trust ratio = 1, which turns this
            into Adam. Useful for comparison purposes.
        _Large Batch Optimization for Deep Learning: Training BERT in 76
            minutes:
        https://arxiv.org/abs/1904.00962
    """

    def __init__(
        self,
        params,
        lr=1e-3,
        betas=(0.9, 0.999),
        eps=1e-6,
        weight_decay=0,
        adam=False,
        min_trust=None,
    ):
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= eps:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")
        if min_trust and not 0.0 <= min_trust < 1.0:
            raise ValueError(f"Minimum trust range from 0 to 1: {min_trust}")
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        self.adam = adam
        self.min_trust = min_trust
        super(Lamb, self).__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    err_msg = (
                        "Lamb does not support sparse gradients, "
                        + "consider SparseAdam instad."
                    )
                    raise RuntimeError(err_msg)

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # Decay the first and second moment running average coefficient
                # m_t
                exp_avg.mul_(beta1).add_((1 - beta1) * grad)
                # v_t
                # exp_avg_sq.mul_(beta2).addcmul_((1 - beta2) * grad *
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=(1 - beta2))

                # Paper v3 does not use debiasing.
                # bias_correction1 = 1 - beta1 ** state['step']
                # bias_correction2 = 1 - beta2 ** state['step']
                # Apply bias to lr to avoid broadcast.
                step_size = group[
                    "lr"
                ]  # * math.sqrt(bias_correction2) / bias_correction1

                weight_norm = p.data.pow(2).sum().sqrt().clamp(0, 10)

                adam_step = exp_avg / exp_avg_sq.sqrt().add(group["eps"])
                if group["weight_decay"] != 0:
                    adam_step.add_(group["weight_decay"], p.data)

                adam_norm = adam_step.pow(2).sum().sqrt()
                if weight_norm == 0 or adam_norm == 0:
                    trust_ratio = 1
                else:
                    trust_ratio = weight_norm / adam_norm
                if self.min_trust:
                    trust_ratio = max(trust_ratio, self.min_trust)
                state["weight_norm"] = weight_norm
                state["adam_norm"] = adam_norm
                state["trust_ratio"] = trust_ratio
                if self.adam:
                    trust_ratio = 1

                p.data.add_(-step_size * trust_ratio * adam_step)

        return loss


class Lookahead(Optimizer):
    def __init__(self, base_optimizer, alpha=0.5, k=6):
        if not 0.0 <= alpha <= 1.0:
            raise ValueError(f"Invalid slow update rate: {alpha}")
        if not 1 <= k:
            raise ValueError(f"Invalid lookahead steps: {k}")
        defaults = dict(lookahead_alpha=alpha, lookahead_k=k, lookahead_step=0)
        self.base_optimizer = base_optimizer
        self.param_groups = self.base_optimizer.param_groups
        self.defaults = base_optimizer.defaults
        self.defaults.update(defaults)
        self.state = defaultdict(dict)
        # manually add our defaults to the param groups
        for name, default in defaults.items():
            for group in self.param_groups:
                group.setdefault(name, default)

    def update_slow(self, group):
        for fast_p in group["params"]:
            if fast_p.grad is None:
                continue
            param_state = self.state[fast_p]
            if "slow_buffer" not in param_state:
                param_state["slow_buffer"] = torch.empty_like(fast_p.data)
                param_state["slow_buffer"].copy_(fast_p.data)
            slow = param_state["slow_buffer"]
            slow.add_(group["lookahead_alpha"] * (fast_p.data - slow))
            fast_p.data.copy_(slow)

    def sync_lookahead(self):
        for group in self.param_groups:
            self.update_slow(group)

    def step(self, closure=None):
        # assert id(self.param_groups) == id(self.base_optimizer.param_groups)
        loss = self.base_optimizer.step(closure)
        for group in self.param_groups:
            group["lookahead_step"] += 1
            if group["lookahead_step"] % group["lookahead_k"] == 0:
                self.update_slow(group)
        return loss

    def state_dict(self):
        fast_state_dict = self.base_optimizer.state_dict()
        slow_state = {
            (id(k) if isinstance(k, torch.Tensor) else k): v
            for k, v in self.state.items()
        }
        fast_state = fast_state_dict["state"]
        param_groups = fast_state_dict["param_groups"]
        return {
            "state": fast_state,
            "slow_state": slow_state,
            "param_groups": param_groups,
        }

    def load_state_dict(self, state_dict):
        fast_state_dict = {
            "state": state_dict["state"],
            "param_groups": state_dict["param_groups"],
        }
        self.base_optimizer.load_state_dict(fast_state_dict)

        # We want to restore the slow state, but share param_groups reference
        # with base_optimizer. This is a bit redundant but least code
        slow_state_new = False
        if "slow_state" not in state_dict:
            print("Loading state_dict from optimizer without Lookahead applied.")
            state_dict["slow_state"] = defaultdict(dict)
            slow_state_new = True
        slow_state_dict = {
            "state": state_dict["slow_state"],
            "param_groups": state_dict[
                "param_groups"
            ],  # this is pointless but saves code
        }
        super(Lookahead, self).load_state_dict(slow_state_dict)
        self.param_groups = (
            self.base_optimizer.param_groups
        )  # make both ref same container
        if slow_state_new:
            # reapply defaults to catch missing lookahead specific ones
            for name, default in self.defaults.items():
                for group in self.param_groups:
                    group.setdefault(name, default)


# %%
if __name__ == "__main__":
    os.makedirs(fig_dir, exist_ok=True)



import numpy as np
import pandas as pd
import torch
from torch import nn

import matplotlib.pyplot as plt
from time import time
from crabnet.utils.get_compute_device import get_compute_device
from crabnet.utils.utils import EDM_CsvLoader

from crabnet.kingcrab import CrabNet
from crabnet.model import Model

compute_device = get_compute_device(prefer_last=True)
print(f"Running on compute device: {compute_device}")


# %%
RNG_SEED = 42
torch.manual_seed(RNG_SEED)
np.random.seed(RNG_SEED)
data_type = torch.float


# %%
# fmt: off
all_symbols = ['None', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na',
               'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc',
               'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga',
               'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb',
               'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb',
               'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
               'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',
               'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl',
               'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa',
               'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
               'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg',
               'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']
# fmt: on

classification = False
batch_size = None
transfer = None
verbose = True
# Get the TorchedCrabNet architecture loaded
mat_prop = "steels_yield"
file_name = rf"data\matbench_cv\{mat_prop}\val0.csv"

model = Model(
    CrabNet(compute_device=compute_device).to(compute_device),
    model_name=f"{mat_prop}",
    verbose=verbose,
)

batch_size = 1
edm_loader = EDM_CsvLoader(csv_data=file_name, batch_size=batch_size)
data_loader = edm_loader.get_data_loaders(inference=True)


# %%
class Embedder(nn.Module):
    def __init__(self, vocab_size, d_model):
        super().__init__()
        self.embed = nn.Embedding(vocab_size, d_model)
        self.mat2vec = True
        if self.mat2vec:
            mat2vec = "data/element_properties/mat2vec.csv"
            # mat2vec = 'data/element_properties/onehot.csv'
            # mat2vec = 'data/element_properties/oliynyk.csv'
            cbfv = pd.read_csv(mat2vec, index_col=0).values
            feat_size = cbfv.shape[-1]
            self.fc_mat2vec = nn.Linear(feat_size, d_model)
            self.fc_mat2vec = self.fc_mat2vec.to(compute_device, dtype=torch.float)
            zeros = np.zeros((1, feat_size))
            cat_array = np.concatenate([zeros, cbfv])
            self.cbfv = torch.tensor(cat_array).to(compute_device, dtype=torch.float)

    def forward(self, src):
        mat2vec = self.cbfv[src]
        # x_emb = mat2vec
        x_emb = self.fc_mat2vec(mat2vec)
        return x_emb


class FractionalEncoder(nn.Module):
    """
    Encoding element fractions using encoding sytle a la Vaswani 2017.
    Unsqueezed the fractional encoding a bit so that we are using more space
    in the d_model dimension to encode fractional information.
    """

    def __init__(self, d_model, resolution=100, log10=False, compute_device=None):
        super().__init__()
        self.d_model = d_model // 2
        self.resolution = resolution
        self.log10 = log10
        self.compute_device = compute_device

        x = torch.linspace(
            0, self.resolution - 1, self.resolution, requires_grad=False
        ).view(self.resolution, 1)
        fraction = (
            torch.linspace(0, self.d_model - 1, self.d_model, requires_grad=False)
            .view(1, self.d_model)
            .repeat(self.resolution, 1)
        )

        pe = torch.zeros(self.resolution, self.d_model)
        pe[:, 0::2] = torch.sin(x / torch.pow(50, 2 * fraction[:, 0::2] / self.d_model))
        pe[:, 1::2] = torch.cos(x / torch.pow(50, 2 * fraction[:, 1::2] / self.d_model))
        pe = self.register_buffer("pe", pe)

    def forward(self, x):
        x = x.clone()
        if self.log10:
            x = 0.0025 * (torch.log2(x)) ** 2
            x[x > 1] = 1
            # x = 1 - x  # for sinusoidal encoding at x=0
        x[x < 1 / self.resolution] = 1 / self.resolution
        frac_idx = torch.round(x * (self.resolution)).to(dtype=torch.long) - 1
        out = self.pe[frac_idx]

        return out


# %%
ti = time()
for i, data in enumerate(data_loader):
    X, y, formula = data
    src, frac = X.squeeze(-1).chunk(2, dim=1)
    src = src.to(compute_device, dtype=torch.long, non_blocking=True)
    frac = frac.to(compute_device, dtype=data_type, non_blocking=True)
    y = y.to(compute_device, dtype=data_type, non_blocking=True)

X, y, formula = data_loader.dataset[23]
src, frac = X.chunk(2, dim=0)
src = src.to(compute_device, dtype=torch.long, non_blocking=True)
frac = frac.to(compute_device, dtype=data_type, non_blocking=True)
y = y.to(compute_device, dtype=data_type, non_blocking=True)
print(i, src)
print(i, frac)
print(src.shape)
print(src.shape)

src[:, :] = 0
src[0, :] = 8
src[1, :] = 13

frac[:, :] = 0
frac[0, :] = 0.6
frac[1, :] = 0.4

vocab_size = 118
d_model = 512

embedder = model.model.encoder.embed
pe = model.model.encoder.pe
ple = model.model.encoder.ple


x = embedder(src)
x = x.masked_fill(src.unsqueeze(1).repeat(1, 1, x.shape[-1]) == 0, 0)

pe_tensor = torch.zeros_like(x)
ple_tensor = torch.zeros_like(x)

pe_tensor[:, :, : d_model // 2] = pe(frac)
ple_tensor[:, :, d_model // 2 :] = ple(frac)

x_cpu = x.cpu().detach().numpy()
pe_tensor_cpu = pe(frac).cpu().detach().numpy()
ple_tensor_cpu = ple(frac).cpu().detach().numpy()

print(x_cpu.shape)
print(pe_tensor_cpu.shape)
print(ple_tensor_cpu.shape)

x_cpu = x_cpu.squeeze(1)
pe_tensor_cpu = pe_tensor_cpu.squeeze(1)
ple_tensor_cpu = ple_tensor_cpu.squeeze(1)

print(x_cpu.shape)
print(pe_tensor_cpu.shape)
print(ple_tensor_cpu.shape)
frac = frac.squeeze(1)
src = src.squeeze(1)


# %%
# colors = ['#7fc97f', '#beaed4', '#fdc086', 'r', 'g', 'steelblue']
colors = [
    "#8dd3c7",
    "#ffffb3",
    "#bebada",
    "#fb8072",
    "#80b1d3",
    "#fdb462",
    "#b3de69",
    "#fccde5",
    "#d9d9d9",
    "#bc80bd",
    "#ccebc5",
    "#ffed6f",
]
nrows = len(src)


fig, axes = plt.subplots(nrows, 1, sharex=True, sharey=True, figsize=(8, 8))
for i, f in enumerate(src):
    element = all_symbols[src[i]]
    axes[i].plot(x_cpu[i], label=f"{element}", color=colors[i])
    axes[i].legend(loc=1, framealpha=0.4, handlelength=0)
    axes[i].set_yticklabels([])
    axes[i].set_facecolor("#202020")
plt.tick_params(which="major", right=True, top=True, direction="in", length=6)
plt.tick_params(which="minor", right=True, top=True, direction="in", length=4)
plt.show()


fig, axes = plt.subplots(nrows, 1, sharex=True, sharey=True, figsize=(8, 8))
for i, f in enumerate(frac):
    axes[i].plot(pe_tensor_cpu[i], label=f"{f:0.3f}", color=colors[i])
    axes[i].legend(loc=1, framealpha=0.4, handlelength=0)
    axes[i].set_yticklabels([])
    axes[i].set_facecolor("#202020")
plt.tick_params(which="major", right=True, top=True, direction="in", length=6)
plt.tick_params(which="minor", right=True, top=True, direction="in", length=4)
plt.show()


fig, axes = plt.subplots(nrows, 1, sharex=True, sharey=True, figsize=(8, 8))
for i, f in enumerate(frac):
    axes[i].plot(ple_tensor_cpu[i], label=f"{f:0.0e}", color=colors[i])
    axes[i].legend(loc=1, framealpha=0.4, handlelength=0)
    axes[i].set_yticklabels([])
    axes[i].set_facecolor("#202020")
plt.tick_params(which="major", right=True, top=True, direction="in", length=6)
plt.tick_params(which="minor", right=True, top=True, direction="in", length=4)
plt.show()


fig, axes = plt.subplots(nrows, 1, sharex=True, sharey=True, figsize=(8, 8))
for i, f in enumerate(frac):
    element = all_symbols[src[i]]
    ebfv = (
        x_cpu[i] * 2
        + pe_tensor[i].cpu().detach().numpy()
        + ple_tensor[i].cpu().detach().numpy()
    )
    ebfv = ebfv.ravel()
    if f == 0:
        ebfv = [0] * len(ebfv)
    axes[i].plot(ebfv, label=f"${element}_{{{f:0.1f}}}$", color=colors[i])
    axes[i].legend(loc=4, framealpha=0.6, handlelength=0)
    axes[i].set_yticklabels([])
    axes[i].set_facecolor("#202020")
plt.tick_params(which="major", right=True, top=True, direction="in", length=6)
plt.tick_params(which="minor", right=True, top=True, direction="in", length=4)
plt.show()



"""Touch up the conda recipe from grayskull using conda-souschef."""
import os
from os.path import join

from souschef.recipe import Recipe

import crabnet

os.system("grayskull pypi {0}=={1}".format(crabnet.__name__, crabnet.__version__))

fpath = join(crabnet.__name__, "meta.yaml")
fpath2 = join("scratch", "meta.yaml")
my_recipe = Recipe(load_file=fpath)
my_recipe["requirements"]["host"].append("flit")
my_recipe["requirements"]["run"].append("pytorch >=1.9.0")
my_recipe["requirements"]["run"].append("cudatoolkit <11.4")
my_recipe.save(fpath)
my_recipe.save(fpath2)



import os
import numpy as np
import warnings
import json

from time import time
from datetime import datetime

from sklearn.ensemble import RandomForestRegressor

from crabnet.utils.get_core_count import get_core_count
from crabnet.utils.modelselectionhelper import modelselectionhelper
from crabnet.utils.utils import NumpyEncoder, CONSTANTS, count_gs_param_combinations


# %%
cons = CONSTANTS()
mat_props_dir = r"data/matbench_cv/"
mat_props = cons.matbench_props

mat_props_names = cons.matbench_names
mat_props_pretty = cons.matbench_names_dict
elem_props = cons.eps
elem_props = ["magpie"]

RNG_SEED = 42
np.random.seed(RNG_SEED)


# %%
models1 = {
    "RandomForestRegressor": RandomForestRegressor(),
}

rf_estimators_range = [500]

params1 = {
    "RandomForestRegressor": {"n_estimators": rf_estimators_range, "max_depth": [None]},
}

scorings = {
    "r2": "r2",
    "neg_MAE": "neg_mean_absolute_error",
    "neg_RMSE": "neg_root_mean_squared_error",
}


# %%
if __name__ == "__main__":
    start_datetime_matbench_classics = datetime.now().strftime("%Y-%m-%d-%H%M%S.%f")

    metrics_dir = f"metrics/rf_gridsearch/"
    fig_dir = r"figures/GridSearchCV/matbench/"
    os.makedirs(metrics_dir, exist_ok=True)
    os.makedirs(fig_dir, exist_ok=True)

    _, cnt_dict = count_gs_param_combinations(params1)
    print(f"Number of parameter combinations for each estimator:\n" f"{cnt_dict}")

    cnt_dict_filename = "parameter_combos_matbench.json"
    cnt_dict_file = os.path.join(metrics_dir, cnt_dict_filename)
    json_content = json.dumps(cnt_dict, cls=NumpyEncoder, indent=4)

    with open(cnt_dict_file, "w") as f:
        try:
            f.write(json_content)
        except:
            pass

    n_cores = get_core_count()
    n_jobs = n_cores // 2 - 2

    ignore_warnings = True
    if ignore_warnings:
        maxiter_msg = (
            "Maximum number of iteration reached "
            "before convergence. Consider increasing max_iter "
            "to improve the fit."
        )
        warnings.filterwarnings("ignore", message=maxiter_msg)

    ti_matbench_classics = time()

    cv_folds = 2
    mshelper1 = modelselectionhelper(
        models1,
        params1,
        elem_props,
        mat_props_dir,
        mat_props,
        metrics_dir,
        fig_dir,
        scoring=scorings,
        n_jobs=n_jobs,
        cv=cv_folds,
        refit="neg_MAE",
        verbose=True,
        random_seed=RNG_SEED,
    )
    dt_matbench_classics = time() - ti_matbench_classics

    print("*********** matbench_classics finished ***********")
    print(
        f"matbench_classics finished, elapsed time: " f"{dt_matbench_classics:0.4g} s"
    )
    print("*********** matbench_classics finished ***********")
